package org.steveshipway.blueprints.directive;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.Logger;

import net.minecraft.block.Block;
import net.minecraft.block.BlockChest;
import net.minecraft.init.Blocks;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.world.World;







//Minecraft 1.7
//import org.steveshipway.blueprints.util17.*;
//Minecraft 1.8
import org.steveshipway.blueprints.util18.*;

import net.minecraft.util.BlockPos;
import net.minecraft.util.WeightedRandomChestContent;
import net.minecraft.block.state.IBlockState;
import net.minecraftforge.common.ChestGenHooks;
import net.minecraftforge.common.util.Constants;

class ChestItem {
	public String id = "minecraft:stick";
	public int mincount = 1;
	public int maxcount = 1;
	public int dv = 0;
	public int chance = 100;
	public NBTTagCompound tag = null;
	public Item item = null;
	public ChestItem(String s, NBTTagCompound itemtag, int d, int n1, int n2, int c) {
		id = s;
		this.tag = itemtag;
		mincount = n1;
		maxcount = n2;
		chance = c;
		if( chance < 0 ) { chance = 0; }
		if( chance > 100 ) { chance = 100; }
		if( mincount < 0) { mincount = 0; }
		if( maxcount < mincount ) { maxcount = mincount; }
		if( mincount > 64 ) { mincount = 64; }
		if( maxcount > 64 ) { maxcount = 64; }
		dv = d & 15;
		this.item = Util.getItemByNameOrId(s);
	}
}
public class ChestDirective extends Directive {
	private BlockPos offset;
	private int facing = 0; // facing direction
	private String contents; // for print/astext only
	private NBTTagCompound nbt; // pre-parsed NBT; if null, use list
	private LinkedList<ChestItem> items; // list of items with probabilities
	private boolean haserror = false;
	private String chestGenCategory = null;
	private int chestGenQty = 0;
	static Pattern itempattern = Pattern.compile("^\\s*(\\S+?)(?:\\((\\d+)\\))?(?:,(\\d+))?(?:,(\\d+))?(?:;(\\d+))?\\s*$");
	static Pattern nbtpattern = Pattern.compile("(?i)^(?:\\s*chest)?\\s*(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(\\d)\\s*(\\S+\\.nbt|\\{.*\\})\\s*$");
	static Pattern listpattern = Pattern.compile("(?i)^(?:\\s*chest)?\\s*(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(\\d)\\s*(.*)\\s*$");
	static Pattern hookpattern = Pattern.compile("(?i)^(?:\\s*chest)?\\s*(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(\\d)\\s*ChestGenHooks?:([a-zA-Z]+),(\\d+)\\s*$");
	
	public ChestDirective(String definition, String bppath,Logger logger) {
		NBTTagCompound nbtItem = null;
		String item = null;
		int dv,minqty,maxqty,chance;
		Matcher m;
		// parse
		// contents can be filename.nbt, raw NBT, or list of ( itemname[,n[,m]][;prob], or itemfile.nbt[,n[,m]][;prob] )
		// turn this into either static nbt, or item list with each item a name+prob+qtyrange+optionaltag
		m = nbtpattern.matcher(definition);
		if ( m.find() ) {
			//logger.info("Chest definition matched explicit NBT: "+definition);
			try {
				this.offset = new BlockPos(Integer.parseInt(m.group(1)),Integer.parseInt(m.group(2)),Integer.parseInt(m.group(3)));
				this.facing = Integer.parseInt(m.group(4));
			} catch ( Exception e ) {
				logger.error("Invalid offset format in CHEST definition '"+definition+"'");
				this.haserror = true;
				return;
			}
			this.nbt = Util.parseNBT(m.group(5),bppath,logger);
			if( this.nbt == null ) {
				logger.error("Unparseable NBT in CHEST definition '"+definition+"'");
				this.haserror = true;
				return;				
			}
			if( this.nbt.getString("id") == null ) {
				this.nbt.setString("id","Chest");
			} else if( this.nbt.getString("id").length()<1 ) {
					this.nbt.setString("id","Chest");
			} else if( !this.nbt.getString("id").contentEquals("Chest") ) {
				logger.error("Supplied NBT in '"+m.group(5)+"' is not for a Chest (id="+this.nbt.getString("id")+")");
				this.haserror = true;
			}
			return;
		}
		m = hookpattern.matcher(definition);
		if( m.find() ) {
			try {
				this.offset = new BlockPos(Integer.parseInt(m.group(1)),Integer.parseInt(m.group(2)),Integer.parseInt(m.group(3)));
				this.facing = Integer.parseInt(m.group(4));
			} catch ( Exception e ) {
				logger.error("Invalid offset format in CHEST definition '"+definition+"'");
				this.haserror = true;
				return;
			}
			this.contents = "ChestGenHooks:"+m.group(5)+","+m.group(6);
			this.chestGenCategory = m.group(5);
			this.chestGenQty = Integer.parseInt(m.group(6));
			this.items = new LinkedList<ChestItem>(); // initially empty
			return;
		}
		m = listpattern.matcher(definition);
		if(! m.find() ) {
			logger.error("Invalid format in CHEST definition '"+definition+"'");
			this.haserror = true;
			return;			
		}
		try {
			this.offset = new BlockPos(Integer.parseInt(m.group(1)),Integer.parseInt(m.group(2)),Integer.parseInt(m.group(3)));
			this.facing = Integer.parseInt(m.group(4));
		} catch ( Exception e ) {
			logger.error("Invalid offset format in CHEST definition '"+definition+"'");
			this.haserror = true;
			return;
		}
		this.contents = m.group(5);
		this.items = new LinkedList<ChestItem>();
		//logger.info("Chest definition matched content list: "+this.contents);
		// is a list of items
		for(String itemdef : this.contents.split("\\s+") ) {
			NBTTagCompound tag = null;
			//CharSequence itemdefcs = itemdef;
			if( itemdef.equals("") ) { continue; }

			m = itempattern.matcher( itemdef );
			if( m.find() ) {
				item = m.group(1);
				//logger.info("Chest item matched pattern: "+item);
				minqty = 1; maxqty = 1; chance = 100;
				dv = 0;
				try {
					if( m.group(3)!= null ) {
						minqty = Integer.parseInt(m.group(3));
					} else { minqty=1; }
					if( m.group(4)!= null ) {
						maxqty = Integer.parseInt(m.group(4));
					} else { maxqty=minqty; }
					if( m.group(5)!= null ) {
						chance = Integer.parseInt(m.group(5));
					} else { chance=100; }
				} catch( Exception e ) {
					logger.error("Invalid numbers in item definition '"+itemdef+"'");
					this.haserror = true;
					continue;					
				}
				if(minqty < 0) { minqty = 1; }
				if(chance<0 || chance>100) { chance = 100; }
				if( item.endsWith(".nbt") || item.startsWith("{") ) {
					nbtItem = Util.parseNBT(item,bppath,logger);
					if( nbtItem!=null ) {
						//extract id, dv and tag from the NBT
						item = nbtItem.getString("id");
						if( nbtItem.hasKey("Damage",Constants.NBT.TAG_SHORT) ) {
							dv = nbtItem.getShort("Damage") & 15;
						}
						if( nbtItem.hasKey("tag") ) {
							tag = (NBTTagCompound) nbtItem.getTag("tag").copy();
						}
					} else {
						logger.error("Invalid NBT in item definition '"+item+"'");
						this.haserror = true;
						continue;
					}
				} else {
					if( !item.contains(":") ) { item = "minecraft:"+item; }

					if( Util.getItemByNameOrId(item) == null ) { // check for validity
						logger.error("Invalid item name defined "+item);
						this.haserror = true;
						continue;
					}
				}
				// allow filename.nbt(dv) to override any Damage tag set in the filename.nbt
				if( m.group(2)!= null && (m.group(2).length()>0)) {
					dv = Integer.parseInt( m.group(2) ) & 15;
				}

				this.items.add(new ChestItem(item,tag,dv,minqty,maxqty,chance));
				
			} else {
				logger.error("Invalid chest item definition: '"+itemdef+"'");
				this.haserror = true;
				return;
			}
		}
	}
	
	@Override
	public boolean build(World world, BlockPos location, int r, Logger logger, HashSet softlist) {
		IBlockState blockState = null;
		BlockPos blockPos = new BlockPos(location);
		int dv = 2; 
		boolean rv;
		NBTTagCompound nbtTag = null;
		
		if( world.isRemote ) {
    		logger.error("Block placement called with remote world?  Should never happen.");
			return false;
		}

		// rotate to correct DV for facing
		switch( (r+facing)&3 ) {
			case 1:
				dv = 5;
				break;
			case 2:
				dv = 3;
				break;
			case 3:
				dv = 4;
				break;
		}
		
		// get block position, rotated
		blockPos = rotateAndAdd(this.offset,r,location);

		// set the block to a chest...
		blockState = Util.getState(Block.getBlockFromName("minecraft:chest"),dv);
		rv = Util.setBlockState(world,blockPos,blockState);
		if( !rv ) {
			logger.warn("Failed to place block in the world");		
			return false;
		}
		if(this.haserror) {
			logger.warn("Chest contents definition has errors");
			// return false;
		}
		// now we have the block to set, but need to add the tile entity for the contents
		if( this.nbt != null ) {
			// static NBT pre-parsed
			nbtTag = (NBTTagCompound) this.nbt.copy();
		} else {
			// random list
			nbtTag = chestTag(logger);
		}
		try {
			//Util.setTileEntity(world, blockPos, TileEntity.createAndLoadEntity(nbtTag));
			Util.setTileEntity(world, blockPos, Util.fixChestNbt(logger,nbtTag));
		} catch(Exception e) {
			logger.warn("Unable to create TileEntity for chest, may be illegal NBT: "+e.getMessage());		
			return false;
		}
		if((this.chestGenCategory != null) && (this.chestGenQty>0)) {
			// Call the ChestGen hook
			Util.callChestGen(world,blockPos,this.chestGenCategory,this.chestGenQty,logger);
		}
		
		return true;
	}
	
	@Override
	public String asText() {
		String s = "CHEST ";
		if( this.haserror ) {
			s = "# CHEST ";
		}
		if( offset != null ) {
			s += " " + offset.getX() + "," + offset.getY() + "," + offset.getZ() + " " + facing;
		} else {
			s += " 0,0,0 0";
		}
		if( contents != null ) {
			s += " " + contents;
		}
		return s;
	}
	
	private NBTTagCompound chestTag(Logger logger) {
		// render the this.list as a random NBTTagCompound
		NBTTagCompound nbtTag  = new NBTTagCompound();
		NBTTagCompound nbtItemTag;
		NBTTagCompound nbtItemOpt = null;
		NBTTagList nbtItems = new NBTTagList();
		int slot = 0;
		int stacksize = 1;
		Item i = null;
		ItemStack itemstack = null;
		Random rand = new Random();
		ChestItem ci;

		// predefined NBT
		if( this.nbt != null ) { return this.nbt; }
		
		// everything has these
		nbtTag.setString("id","Chest");
		nbtTag.setString("Lock","");
		
		if( this.items == null ) { //empty chest
			return nbtTag;
		}
		if( this.nbt != null ) {
			return this.nbt;       // preparsed
		}
		// now the items
		Iterator ii = this.items.iterator();
		while( ii.hasNext() ) {
			ci = (ChestItem) ii.next();
			i = Util.getItemByNameOrId(ci.id);
			nbtItemOpt = null;
			if( i == null ) {
				// invalid item!
				logger.warn("Invalid item name in chest contents: "+ci.id);
				continue;
			}
			if( ci.chance == 0 ) { continue; }
			if( rand.nextInt(100) >= ci.chance ) { continue; }
			nbtItemOpt = ci.tag;
			itemstack = new ItemStack(i,ci.dv);
			if( ci.maxcount <= ci.mincount ) {
				stacksize = ci.maxcount;
			} else {
				stacksize = ci.mincount + rand.nextInt(ci.maxcount - ci.mincount);
			}
			if( stacksize > i.getItemStackLimit(itemstack) ) { stacksize = i.getItemStackLimit(itemstack); }
			if( stacksize < 1 ) { continue; } // skip this item as we have 0
			// now add this item to the NBT
			nbtItemTag = new NBTTagCompound();
			nbtItemTag.setByte("Count", (byte) stacksize);
			nbtItemTag.setByte("Slot", (byte) slot);
			nbtItemTag.setShort("Damage", (short) ci.dv);
			nbtItemTag.setString("id", ci.id);
			if( nbtItemOpt != null ) {
				nbtItemTag.setTag("tag",nbtItemOpt);
			}

			// add this itemstack NBT to the Chest's items list
			nbtItems.appendTag(nbtItemTag);			
			slot += 1;
		}
		nbtTag.setTag("Items",nbtItems);			
		
		return nbtTag;
	}
}
