package org.steveshipway.blueprints.directive;

import java.util.HashSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.Logger;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.item.Item;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.world.World;



// Minecraft 1.7
//import org.steveshipway.blueprints.util17.*;
// Minecraft 1.8
import org.steveshipway.blueprints.util18.*;

import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.EntityList;
import net.minecraft.util.BlockPos;

public class SpawnerDirective  extends Directive {
	private String mobdata;
	private NBTTagCompound nbt;
	private BlockPos offset;
	private boolean haserror = false;
	
	static Pattern spawnpattern = Pattern.compile("^\\s*([^\\s;]+)(?:;(\\d+))?\\s*$");
	static Pattern optpattern = Pattern.compile("^\\s*(\\w+):(\\d+)\\s*$");
	static Pattern nbtpattern = Pattern.compile("(?i)^(?:\\s*spawner)?\\s*(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)\\s*(\\S+\\.nbt|\\{.*\\})\\s*$");
	static Pattern listpattern = Pattern.compile("(?i)^(?:\\s*spawner)?\\s*(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)\\s*(.*)$");
	
	public SpawnerDirective(String definition, String bppath,Logger logger) {
		// parse options
		// SPAWNER x y z filename.nbt
		// SPAWNER x y z { nbt }
		// SPAWNER x y z [count[,range[,minticks[,maxticks]]]] potential[;weight] [potential[;weight]] ...
		NBTTagCompound nbtPotential = null;
		NBTTagCompound nbtTag = new NBTTagCompound();
		NBTTagList nbtPList;
		String potential = null;
		int chance;
		Matcher m;
		int seq = 1;
		
		this.mobdata = definition; // for use by print/save only
		
		// parse explicit NBT for the spawner
		m = nbtpattern.matcher(definition);
		if ( m.find() ) {
			//logger.info("Spawner def matches explicit NBT pattern: "+definition);
			try {
				this.offset = new BlockPos(Integer.parseInt(m.group(1)),Integer.parseInt(m.group(2)),Integer.parseInt(m.group(3)));
			} catch ( Exception e ) {
				logger.error("Invalid offset format in SPAWNER definition '"+definition+"'");
				this.haserror = true;
				return;
			}
			this.nbt = Util.parseNBT(m.group(4),bppath,logger);
			if( this.nbt == null ) {
				logger.error("Invalid NBT in SPAWNER definition '"+definition+"'");
				this.haserror = true;
				return;				
			}
			if( this.nbt.getString("id") == null ) {
				this.nbt.setString("id","MobSpawner");
			} else if( this.nbt.getString("id").length()<1 ) {
				this.nbt.setString("id","MobSpawner");
			} else if(! this.nbt.getString("id").contentEquals("MobSpawner") ) {
				logger.error("Supplied NBT in "+m.group(4)+" is not for a MobSpawner (id="+this.nbt.getString("id")+")");
				this.haserror = true;
			}
			return;
		}
		m = listpattern.matcher(definition);
		if(! m.find() ) {
			logger.error("Invalid format in SPAWNER definition '"+definition+"'");
			this.haserror = true;
			return;
		}
		try {
			this.offset = new BlockPos(Integer.parseInt(m.group(1)),Integer.parseInt(m.group(2)),Integer.parseInt(m.group(3)));
		} catch ( Exception e ) {
			logger.error("Invalid offset format in SPAWNER definition '"+definition+"'");
			this.haserror = true;
			return;
		}
		this.mobdata = m.group(4);
		//logger.info("Spawner def matches individual potentials pattern: "+definition);
		// is a list of potentials
		// Set base options
		nbtTag.setString("id","MobSpawner");
		nbtTag.setShort("Delay",(short) -1); // force a reset
		nbtTag.setShort("SpawnCount",(short) 1);
		nbtTag.setShort("SpawnRange",(short) 4);
		nbtTag.setShort("MinSpawnDelay",(short) 100);
		nbtTag.setShort("MaxSpawnDelay",(short) 200);
		nbtPList = new NBTTagList();

		for(String potentialdef : this.mobdata.split("\\s+") ) {
			if( potentialdef.equals("") ) { continue; }
			// parse options definition
			m = optpattern.matcher( potentialdef );
			if( m.find() ) {
				// let's assume they know what they're doing (hahahahaHAHAHA)
				// any potential of the form   Foobar:123   will set spawner option Foobar to value 123
				nbtTag.setShort(m.group(1),(short)Integer.parseInt(m.group(2)));
				continue;
			}
			
			// parse potential definition
			m = spawnpattern.matcher( potentialdef );
			if( m.find() ) {
				potential = m.group(1);
				//logger.info("Spawner potential ("+potential+") matches pattern: "+potentialdef);
				if( m.group(2)!= null ) {
					chance = Integer.parseInt(m.group(2));
				} else { chance=100; }
				if(chance<0) { chance = 100; }
				nbtPotential = new NBTTagCompound();
				nbtPotential.setInteger("Weight",chance);
				if( potential.endsWith(".nbt") || potential.startsWith("{") ) {
					NBTTagCompound tag = Util.parseNBT(potential,bppath,logger);
					if( tag!=null ) {
						//extract id and tag from the NBT
						NBTBase ptag;
						if( tag.getString("id") != null && tag.getString("id").length()>0) {
							nbtPotential.setString("Type", tag.getString("id"));
							tag.removeTag("id");
							nbtPotential.setTag("Properties",tag.copy());
						} else if( tag.getString("Type") != null && tag.getString("Type").length()>0 ) {
							nbtPotential.setString("Type", tag.getString("Type"));
							ptag = tag.getTag("Properties");
							if(ptag != null ) {
								nbtPotential.setTag("Properties",ptag.copy());
							} else {
								nbtPotential.setTag("Properties",new NBTTagCompound());								
							}
							//logger.info("Spawner potential - NBT with Type");
						} else if( tag.getString("EntityId") != null && tag.getString("EntityId").length()>0) {
							nbtPotential.setString("Type", tag.getString("EntityId"));
							ptag = tag.getTag("SpawnData");
							if(ptag != null ) {
								nbtPotential.setTag("Properties",ptag.copy());
							} else {
								nbtPotential.setTag("Properties",new NBTTagCompound());								
							}
							//logger.info("Spawner potential - NBT with EntityId");
						} else {
							logger.error("Invalid NBT data '"+potential+"' does not have Type or EntityId");
							this.haserror = true;							
							continue;
						}
					} else {
						logger.error("Invalid NBT data in '"+potential+"'");
						this.haserror = true;
						continue;
					}
				} else {
					if( Util.isValidEntityName(potential) ) {
						nbtPotential.setString("Type",potential);
						nbtPotential.setTag("Properties",new NBTTagCompound());
						//logger.info("Spawner potential - Type="+potential);
					} else {
						logger.error("Invalid Entity name in SPAWNER definition: '"+potential+"'");
						this.haserror = true;
						continue;						
					}
					
				}
			} else {
				logger.error("Invalid spawner potential definition: '"+potentialdef+"'");
				this.haserror = true;
				continue;
			}
			// next item in list...
			nbtPList.appendTag(nbtPotential);
			seq = seq + 1;
		}
		nbtTag.setTag("SpawnPotentials",nbtPList);
		if( nbtPList.tagCount() > 0 ) {
			try {
				NBTBase firstPotential =  Util.getIndex(nbtPList, 0);
				
				nbtTag.setString("EntityId", ((NBTTagCompound)firstPotential).getString("Type") );
				nbtTag.setTag("SpawnData", ((NBTTagCompound)firstPotential).getTag("Properties").copy() );
			} catch( Exception e ) {
				// let it go...
				logger.warn("Def: "+mobdata);
				logger.warn("Possible problem in duplicating primary spawn potential to base definition");
				logger.warn("Error: "+e.toString());
			}
		}
		this.nbt = Util.fixSpawnerNbt(logger, nbtTag);
	}
	
	@Override
	public boolean build(World world, BlockPos location, int r, Logger logger, HashSet softlist) {
		IBlockState blockState = null;
		BlockPos blockPos = new BlockPos(location); 
		boolean rv;
		NBTTagCompound nbtTag = null;
		
		if( world.isRemote ) {
    		logger.error("Block placement called with remote world?  Should never happen.");
			return false;
		}
		
		// get block position, rotated
		blockPos = rotateAndAdd(this.offset,r,location);

		// set the block to a spawner...
		rv = Util.setBlockState(world,blockPos,Util.getState(Blocks.mob_spawner,0));
		if( !rv ) {
			logger.warn("Failed to place spawner in the world at "+blockPos.getX()+","+blockPos.getY()+","+blockPos.getZ());		
			return false;
		}
		if(this.haserror || this.nbt==null) {
			logger.warn("Spawner definition has errors");
			return false;
		}
		// now we have the block to set, but need to add the tile entity for the contents
		try {
			Util.setTileEntity(world, blockPos,TileEntity.createAndLoadEntity(this.nbt));
		} catch(Exception e) {
			logger.warn("Unable to create TileEntity for spawner, may be illegal NBT: "+e.getMessage());		
			return false;
		}
		
		return true;
	}
	
	@Override
	public String asText() {
		String s = "SPAWNER ";
		if( this.haserror || (this.nbt == null)) {
			return "# SPAWNER undefined ";
		}
		s += " " + offset.getX() + "," + offset.getY() + "," + offset.getZ();
		if( this.mobdata != null ) {
			s += " " + this.mobdata;
		}
		return s;
	}
}
